import socket
import time
from datetime import datetime
from multiprocessing import Process

import psutil
import requests
import setproctitle

from distributed_computing.common import RedisQueue
from distributed_computing.client import PoolClient
from distributed_computing.globals import DEFAULT_PORT, MANAGEMENT_Q_NAME, SERVER_PROCESS_NAME, CLIENT_PROCESS_NAME
from distributed_computing.server import PoolServer


def _run_pool_client(host, port, password, max_gpus):
    setproctitle.setproctitle(CLIENT_PROCESS_NAME)
    client = PoolClient(host, port, password, max_gpus)
    client.start()


def start_node(password, address='localhost', port=DEFAULT_PORT, max_gpus=None):
    """
    Start a worker node on the local machine.
    :param password: worker pool password to prevent unauthorized access.
    :param address: workers pool server address. If not provided, assumed to be the local machine (localhost).
    :param port: workers pool server port.
    :param violent_exit: Use sys.exit(1) to kill the node when SIGKILL is triggered.
    :return: a function that terminates the node process. A function that waits for it to finish.
    """
    if is_client_running():
        print('A client is already running on this machine.')
        return lambda: 0, lambda: 0

    address = address or 'localhost'

    print('Starting workers pool client.')

    client = Process(target=_run_pool_client, args=(address, port, password, max_gpus))
    client.start()

    return lambda: client.terminate(), lambda: client.join()


def _run_pool_server(host, port, password):
    setproctitle.setproctitle(SERVER_PROCESS_NAME)
    server = PoolServer(host, port, password)
    server.start()


def start_server(password, port=DEFAULT_PORT):
    """
    Start the worker pool server on the local machine. May be used to separate pool management from worker machines.
    A redis server is also started by this function if not already running.
    :param password: worker pool password to prevent unauthorized access.
    :param port: workers pool server port.
    :return: a function that terminates the pool server process. A function that waits for it to finish.
    """
    if is_server_running():
        print('A server is already running on this machine.')
        return lambda: 0, lambda: 0

    print('Starting workers pool server.')
    pool_server = Process(target=_run_pool_server, args=('localhost', port, password))
    pool_server.start()

    print('\n' + '\033[36m*\033[0m' * 90)
    print('\033[1mTo start a worker node, type:\033[0m')
    ip = requests.get('https://api.ipify.org').text

    print(f'\033[92mstart_node -p {port} -a {ip} -r {password}\033[0m')
    print('\033[36m*\033[0m' * 90 + '\n')

    return lambda: pool_server.terminate(), lambda: pool_server.join()


def start_head_node(password, port=DEFAULT_PORT, max_gpus=None):
    """
    Start a worker pool server and a worker node on the local machine. A redis server is also started by this function
    if not already running.
    :param password: worker pool password to prevent unauthorized access.
    :param port: workers pool server port.
    :return: a function that terminates both the server and the node process. A function that waits for both it to
    finish.
    """
    server_terminator, server_waiter = start_server(password, port)
    client_terminator, client_waiter = start_node(password, port=port, max_gpus=max_gpus)

    def terminate():
        server_terminator()
        client_terminator()

    def wait():
        server_waiter()
        client_waiter()

    return terminate, wait


def is_server_running():
    """
    Check if a worker pool server is already running on the local machine.
    :return: True if a server is running, else False
    """
    return SERVER_PROCESS_NAME in [p.name() for p in psutil.process_iter()]


def is_client_running():
    """
    Check if a worker node is already running on the local machine.
    :return: True if a server is running, else False
    """
    return CLIENT_PROCESS_NAME in [p.name() for p in psutil.process_iter()]


class Pool(object):
    """
    Worker pool interface, similar to python multiprocessing Pool interface.
    """
    def __init__(self, worker_class, init_data, password, server_address='localhost', server_port=DEFAULT_PORT,
                 job_timeout=None, min_gpu_memory_required=11000, min_workers=0):
        """
        Initialize pool and connect to the worker pool server.
        :param worker_class: A worker class that implements the WorkerInterface. This class is instantiated on each
        worker process in the worker machines.
        :param init_data: Initialization data that is passed to the worker class constructor. This is a dictionary
        of the form {'args': [...], 'kwargs': {...}}. The args are provided to the worker class constructor as args,
        and the kwargs are provided as keyword args.
        :param password: worker pool password to prevent unauthorized access.
        :param server_address: workers pool server address. If not provided, assumed to be the local machine
        (localhost).
        :param server_port: workers pool server port.
        """
        self._address = server_address
        self._port = server_port
        self._password = password

        self._name = f'{socket.gethostname()}@{worker_class.__name__}'

        self._management_q = RedisQueue(MANAGEMENT_Q_NAME, self._password, self._address, self._port)
        self._private_q = RedisQueue(self._name, self._password, self._address, self._port)

        self._management_q.put(('TASK_ASSIGNMENT', {'init_data': init_data, 'worker_class': worker_class,
                                                    'name': self._name, 'job_timeout': job_timeout,
                                                    'min_gpu_memory_required': min_gpu_memory_required}))
        message, data = self._private_q.get()

        if message != 'TASK_APPROVAL':
            raise Exception('Work was not approved by the pool server')

        self._jobs_q_name = data['jobs_q_name']
        self._results_q_name = data['results_q_name']

        # TODO: what to do when workers are leaving and this number is reached?
        if min_workers:
            self.wait_for_workers(min_workers)

    def imap_unordered(self, data):
        """
        Return results once ready. Order is not guaranteed. Similar to multiprocessing.Pool.imap_unordered.
        :param data: List of items to be processed by the worker. This is a dictionary
        of the form {'args': [...], 'kwargs': {...}}. The args are provided to the worker class run method as args,
        and the kwargs are provided as keyword args.
        """
        for _, result in self._enumerated_imap_unordered(data):
            yield result

    def imap(self, data):
        """
        Return all the results, ordered by the corresponding inputs ordering. Similar to multiprocessing.Pool.map.
        :param data: List of items to be processed by the worker. This is a dictionary
        of the form {'args': [...], 'kwargs': {...}}. The args are provided to the worker class run method as args,
        and the kwargs are provided as keyword args.
        :return: results list.
        """
        results = {}
        returned = 0

        for index, result in self._enumerated_imap_unordered(data):
            results[index] = result

            if returned in results:
                yield results[returned]
                del results[returned]
                returned += 1

    def map(self, data):
        """
        Return all the results, ordered by the corresponding inputs ordering. Similar to multiprocessing.Pool.map.
        :param data: List of items to be processed by the worker. This is a dictionary
        of the form {'args': [...], 'kwargs': {...}}. The args are provided to the worker class run method as args,
        and the kwargs are provided as keyword args.
        :return: results list.
        """
        results = [None] * len(data)

        for index, result in self._enumerated_imap_unordered(data):
            results[index] = result

        return results

    def get_workers_count(self):
        """
        Get the current number of worker processes. Not to be confused with the number of worker nodes (machines).
        :return: number of worker processes.
        """
        self._management_q.put(('COUNTS_REQUEST', None))
        message, data = self._private_q.get()

        assert message == 'COUNTS_RESPONSE', 'Invalid response for WORKERS_COUNTS'

        return data['workers_count']

    def get_nodes_count(self):
        """
        Get the current number of worker nodes. Not to be confused with the number of worker processes.
        :return: number of worker nodes.
        """
        self._management_q.put(('COUNTS_REQUEST', None))
        message, data = self._private_q.get()

        assert message == 'COUNTS_RESPONSE', 'Invalid response for WORKERS_COUNTS'

        return data['clients_count']

    def wait_for_workers(self, min_workers=1, timeout=None):

        start_time = datetime.now()

        while timeout is None or datetime.now() - start_time < timeout:
            if self.get_workers_count() >= min_workers:
                return

            time.sleep(1)

    def update_workers(self, data):
        """
        Send a synchronous update to all worker processes. The workers will get this update once they finish the
        current job, or immediately if no job is being processed.
        :param data: update data that is passed to the worker handle_update method. This is a dictionary
        of the form {'args': [...], 'kwargs': {...}}. The args are provided to the worker class handle_update method as
        args, and the kwargs are provided as keyword args.
        :return: None
        """
        self._management_q.put(('TASK_UPDATE', data))

    def close(self):
        """
        Close the current pool. Active jobs, if any, are aborted.
        :return: None
        """
        pass

    def _enumerated_imap_unordered(self, jobs_data):

        jobs_q = RedisQueue(self._jobs_q_name, self._password, self._address, self._port)
        results_q = RedisQueue(self._results_q_name, self._password, self._address, self._port)

        jobs_data = list(jobs_data)

        # TODO: postpone the later parts addition if the jobs are too heavy. Push them only when the first ones were
        #  finished.
        for i, item in enumerate(jobs_data):
            jobs_q.put((i, item))

        remaining = set(range(len(jobs_data)))

        while len(remaining) > 0:

            index, result = results_q.get()

            if result is None:
                jobs_q.put((index, jobs_data[index]))
                continue

            if index in remaining:
                remaining.remove(index)
                yield index, result

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class WorkerInterface(object):
    """
    This interface must be implemented by the worker class provided to the Pool object.
    """
    def run(self, *args, **kwargs):
        """
        This method is where the main work is done. It is called for each job, with the corresponding args and kwargs
        provided to the pool.map or pool.imap_unordered functions. The returned value is passed to the caller as is.
        """
        raise NotImplemented()

    def handle_update(self, *args, **kwargs):
        """
        This method is called when the pool.update_workers is called by the pool owner with the provided args and
        kwargs. The returned value is ignored.
        """
        raise NotImplemented


import os
import subprocess
import tempfile
import time
from datetime import datetime
import atexit
import redis_server
from redis import Redis

from distributed_computing.common import RedisQueue
from distributed_computing.globals import MANAGEMENT_Q_NAME


class PoolServer(object):

    def __init__(self, host, port, password, keep_updates_history=True):
        self._host = host
        self._port = port
        self._password = password
        self._keep_updates_history = keep_updates_history
        self._updates_history = []
        self._clients = {}
        self._last_heartbeat_by_client = {}
        self._init_data = None
        self._management_q = None
        self._jobs_q_name = None
        self._results_q_name = None
        self._job_timeout = None

    def start(self):

        redis_terminator = self.start_redis(self._port)
        atexit.register(redis_terminator)

        self._management_q = RedisQueue(MANAGEMENT_Q_NAME, self._password, self._host, self._port)

        while True:

            self._check_for_disconnected_clients()

            item = self._management_q.get(timeout=5)

            if item is None:
                continue

            message, data = item

            if message == 'TASK_ASSIGNMENT':
                self._handle_task_assignment(data)

            elif message == 'TASK_UPDATE':
                self._handle_task_update(data)

            elif message == 'WORKER_STATUS_UPDATE':
                self._handle_worker_status_update(data)

            elif message == 'COUNTS_REQUEST':
                self._handle_count_request()

            elif message == 'CLIENT_HEARTBEAT':
                self._handle_client_heartbeat(data)

            else:
                print(f'Unknown message received: {message}. Ignoring.')

    def _check_for_disconnected_clients(self):

        for client in list(self._last_heartbeat_by_client):

            if (datetime.now() - self._last_heartbeat_by_client[client]).total_seconds() > 180:
                print(f'Didn\'t hear from {client} for too long. Rescheduling its tasks.')

                # Reschedule client's active jobs
                for worker, info in self._clients[client]['workers'].items():

                    if info.get('active_job') is not None:
                        print('Rescheduling job', info['active_job'])
                        self._results_q.put((info['active_job'], None))
                        info['active_job'] = None
                        info['job_start_time'] = None

                    info['private_q'].clear()

                # If the the client queue is empty, it might be able to reset its workers.
                if self._clients[client]['private_q'].qsize() == 0:
                    print('Asking it to reset its workers.')
                    self._last_heartbeat_by_client[client] = datetime.now()

                    self._clients[client]['private_q'].put(('TASK_ASSIGNMENT', {
                        'init_data': self._init_data, 'jobs_q_name': self._jobs_q_name,
                        'worker_class': self._worker_class, 'results_q_name': self._results_q_name,
                        'gpu_memory_required': self._min_gpu_memory_required
                    }))

                # The client didn't receive the reset request. Give up.
                else:
                    print('Removing it from the registry')
                    self._last_heartbeat_by_client.pop(client)
                    self._clients.pop(client)

    def _handle_client_heartbeat(self, info):

        if self._init_data is None:
            return

        # A new client
        if info['name'] not in self._clients:
            print('The node', info['name'], 'is joining the pool.')
            self._register_client(info)

        # A rejoining client
        elif info['first'] is True:
            for worker, worker_info in self._clients[info['name']]['workers'].items():
                if worker_info['active_job'] is not None:
                    print('Rescheduling job', worker_info['active_job'], 'of', info['name'])
                    self._results_q.put((worker_info['active_job'], None))
                    worker_info['active_job'] = None
                    worker_info['job_start_time'] = None

            ws = 's' if len(info["worker_names"]) > 1 else ''
            print(f'Rejoining the node {info["name"]} with {len(info["worker_names"])} worker{ws}.')
            self._register_client(info)

        # Just a heartbeat
        else:
            self._last_heartbeat_by_client[info['name']] = datetime.now()

            # Update new workers if there are such
            for worker_name in info['worker_names']:
                if worker_name not in self._clients[info['name']]['workers']:
                    worker_q = RedisQueue(f'{info["name"]}@{worker_name}', self._password, self._host,
                                          self._port).clear()
                    self._clients[info['name']]['workers'][worker_name] = {'active_job': None, 'job_start_time': None,
                                                                           'private_q': worker_q}

            if self._job_timeout is None:
                return

            # Check for stuck workers
            for client_name, client_info in self._clients.items():

                for worker_name in list(client_info['workers']):

                    worker_info = client_info['workers'][worker_name]

                    if worker_info['job_start_time'] is None:
                        continue

                    time_since_job_start = datetime.now() - worker_info['job_start_time']

                    if time_since_job_start.total_seconds() > self._job_timeout:
                        print('The job', worker_info['active_job'], ' of worker', worker_name, 'on client',
                              client_name, f'is taking too long ({time_since_job_start.total_seconds()} seconds). '
                              'Rescheduling and resetting worker.')
                        self._results_q.put((worker_info['active_job'], None))
                        client_info['workers'].pop(worker_name)
                        client_info['private_q'].put(('WORKER_RESET', worker_name))

    def _handle_count_request(self):
        #print({c: [w for w in cc['workers']] for c, cc in self._clients.items()})
        workers_count = sum(len(c['workers']) for c in self._clients.values())
        self._master_q.put(('COUNTS_RESPONSE', {'clients_count': len(self._clients), 'workers_count': workers_count}))

    def _handle_worker_status_update(self, data):

        if data['client'] not in self._clients:
            print(f'Received a status update from an unknown node: {data["client"]}. Ignoring.')
            return

        if data['worker'] not in self._clients[data['client']]['workers']:
            private_q = RedisQueue(f'{data["client"]}@{data["worker"]}', self._password, self._host, self._port).clear()
            self._clients[data['client']]['workers'][data['worker']] = {'private_q': private_q, 'active_job': None,
                                                                        'job_start_time': None}
        if data['status'] == 'in_progress':

            # If the worker took a new job without finishing the previous one, reschedule the previous one.
            if self._clients[data['client']]['workers'][data['worker']]['active_job'] is not None:
                print(f'An unfinished job detect for the node {data["client"]}. Rescheduling.')
                self._results_q.put((self._clients[data['client']]['workers'][data['worker']]['active_job'], None))

            self._clients[data['client']]['workers'][data['worker']]['active_job'] = data['job']
            self._clients[data['client']]['workers'][data['worker']]['job_start_time'] = datetime.now()

        # The job is done
        else:
            self._clients[data['client']]['workers'][data['worker']]['active_job'] = None
            self._clients[data['client']]['workers'][data['worker']]['job_start_time'] = None

        self._last_heartbeat_by_client[data['client']] = datetime.now()

    def _handle_task_update(self, update_data):

        if self._keep_updates_history:
            self._updates_history.append(update_data)

        for client_name, client_info in self._clients.items():
            for worker_name, worker_info in client_info['workers'].items():
                worker_info['private_q'].put(update_data)

    def _handle_task_assignment(self, registration_info):

        self._init_data = registration_info['init_data']
        self._job_timeout = registration_info['job_timeout']
        self._min_gpu_memory_required = registration_info['min_gpu_memory_required']
        self._worker_class = registration_info['worker_class']
        self._jobs_q_name = f'{registration_info["worker_class"].__name__}_jobs'
        self._results_q_name = f'{registration_info["worker_class"].__name__}_results'
        self._results_q = RedisQueue(self._results_q_name, self._password, self._host, self._port)

        RedisQueue(self._jobs_q_name, self._password, self._host, self._port).clear()

        self._master_q = RedisQueue(registration_info['name'], self._password, self._host, self._port)
        self._updates_history = []

        for client_name, client_info in self._clients.items():

            worker_q_names = {worker_name: f'{client_name}@{worker_name}' for worker_name in client_info['workers']}

            client_info['private_q'].put(('TASK_ASSIGNMENT', {
                'init_data': self._init_data, 'jobs_q_name': self._jobs_q_name, 'worker_class': self._worker_class,
                'results_q_name': self._results_q_name, 'worker_q_names': worker_q_names,
                'gpu_memory_required': self._min_gpu_memory_required
            }))

        for client_name, client_info in self._clients.items():
            client_info['private_q'].clear()
            for worker_name, worker_info in client_info['workers'].items():
                worker_info['private_q'].clear()
                worker_info['active_job'] = None
                worker_info['job_start_time'] = None

        time.sleep(10)

        # Assuming that if there were existing workers, they were all killed by now
        self._results_q.clear()

        self._master_q.put(('TASK_APPROVAL', {'jobs_q_name': self._jobs_q_name,
                                              'results_q_name': self._results_q_name}))

    def _register_client(self, registration_info):

        client_q = RedisQueue(registration_info["name"], self._password, self._host, self._port).clear()
        client_info = {'private_q': client_q, 'workers': {}}

        if len(self._updates_history) > 0:
            wp = 's' if len(registration_info['worker_names']) > 1 else ''
            us = 's' if len(self._updates_history) > 1 else ''
            print(f'Updating {registration_info["name"]} worker{wp} with {len(self._updates_history)} update{us}')

        self._clients[registration_info['name']] = client_info
        self._last_heartbeat_by_client[registration_info["name"]] = datetime.now()

        client_q.put(('TASK_ASSIGNMENT', {
            'init_data': self._init_data, 'jobs_q_name': self._jobs_q_name, 'worker_class': self._worker_class,
            'results_q_name': self._results_q_name, 'gpu_memory_required': self._min_gpu_memory_required
        }))

    def start_redis(self, port):
        """
        Start a dedicated redis server on the local machine.
        :param password: worker pool password to prevent unauthorized access.
        :return: None if the redis server is already running, or a process handle if it has been started by this function.
        """

        try:
            Redis(f'localhost:{port}', password=self._password, socket_connect_timeout=1).ping()
            return lambda: None
        except:
            _, conf_file = tempfile.mkstemp()

            with open(conf_file, 'w') as f:
                f.write(f'requirepass {self._password}')

            FNULL = open(os.devnull, 'w')

            print('Redis server is started.')

            p = subprocess.Popen([redis_server.REDIS_SERVER_PATH, conf_file, '--protected-mode no'], stdout=FNULL)

            return lambda: p.terminate()


import atexit
import os
import signal
import socket
import time
from collections import Counter
from multiprocessing import Process

import GPUtil
import psutil

from distributed_computing.common import RedisQueue
from distributed_computing.globals import MANAGEMENT_Q_NAME


class ExperimentWorkerExecutor(Process):

    def __init__(self, name, client_name, gpu_idx, job_q, results_q, master_q, private_q, worker_class, init_data,
                 master_host, master_port, master_password):

        super(ExperimentWorkerExecutor, self).__init__()

        self.name = name
        self.client_name = client_name
        self.jobs_q_name = job_q
        self.results_q_name = results_q
        self.private_q_name = private_q
        self.master_q_name = master_q
        self.gpu_idx = gpu_idx
        self.init_data = init_data
        self.worker_class = worker_class
        self.head_address = master_host
        self.head_password = master_password
        self.master_port = master_port

    def _get_worker_instance(self, init_data):
        """
        The worker class must be loaded *AFTER* the CUDA_VISIBLE_DEVICES environment variable is set.
        So the module is extracted from the class and re-imported.
        """
        __import__(self.worker_class.__module__, fromlist=[self.worker_class.__name__])
        return self.worker_class(*init_data.get('args', []), **init_data.get('kwargs', {}))

    def run(self):

        jobs_q = RedisQueue(self.jobs_q_name, self.head_password, self.head_address)
        results_q = RedisQueue(self.results_q_name, self.head_password, self.head_address)
        private_q = RedisQueue(self.private_q_name, self.head_password, self.head_address)
        master_q = RedisQueue(self.master_q_name, self.head_password, self.head_address)

        # Without this variable, the GPU numbering is different than the one returned by GPUtil/nvidia-smi
        os.environ['CUDA_DEVICE_ORDER'] = 'PCI_BUS_ID'
        os.environ['CUDA_VISIBLE_DEVICES'] = str(self.gpu_idx)
        worker = self._get_worker_instance(self.init_data)

        while True:

            index, data = jobs_q.get()
            master_q.put(('WORKER_STATUS_UPDATE', {'client': self.client_name, 'worker': self.name,
                                                   'job': index, 'status': 'in_progress'}))
            print(f'\033[96m{self.name} >>\033[0m Processing job {index}')

            for update_data in private_q.get_all_nowait():
                worker.handle_update(*update_data.get('args', []), **update_data.get('kwargs', {}))

            result = worker.run(*data.get('args', []), **data.get('kwargs', {}))
            results_q.put((index, result))
            master_q.put(('WORKER_STATUS_UPDATE', {'client': self.client_name, 'worker': self.name,
                                                   'job': index, 'status': 'finished'}))
            print(f'\033[96m{self.name} >>\033[0m Done processing job {index}')


class PoolClient(object):

    def __init__(self, host, port, password, max_gpus):

        self._host = host
        self._port = port
        self._password = password
        self._max_gpus = max_gpus
        self._workers = {}
        self._gpu_memory_required = 0
        self._name = socket.gethostname()
        self._task_assigned = False
        self._management_q = RedisQueue(MANAGEMENT_Q_NAME, self._password, self._host, self._port)

        atexit.register(self._kill_all_workers)

    def start(self):

        private_q = RedisQueue(self._name, self._password, self._host, self._port)
        self._send_heartbeat_to_server(first=True)

        while True:

            item = private_q.get(timeout=10)

            self._send_heartbeat_to_server()

            # if self._task_assigned:
            #     self._start_available_workers()

            if item is None:
                continue

            message, data = item

            if message == 'TASK_ASSIGNMENT':
                self._handle_task_assignment(data)

            elif message == 'WORKER_RESET':
                self._handle_worker_reset(data)

            else:
                print('Ignoring unknown message:', message)

    def _start_available_workers(self):

        for worker_name, info in self._get_available_gpus().items():
            if worker_name in self._workers:
                continue

            if self._max_gpus is None or len(self._workers) < self._max_gpus:
                if info['free_gpu_memory'] > self._gpu_memory_required:
                    self._workers[worker_name] = info
                    self._workers[worker_name]['pid'] = self._start_worker(worker_name, info['gpu_index'])

    def _handle_worker_reset(self, worker_name):

        self._workers[worker_name]['resets_count'] = self._workers[worker_name].get('resets_count', 0) + 1

        if not psutil.pid_exists(self._workers[worker_name]['pid']):
            print('Worker is already dead. Reset aborted.')
            self._workers.pop(worker_name)
            return

        print('Killing worker:', worker_name)
        self._kill_worker(worker_name)

        if self._workers[worker_name]['resets_count'] == 2:
            self._workers.pop(worker_name)
            print('Two reset requests for worker', worker_name, 'were counted. Stopping it permanently.')
            return

        time.sleep(10)

        print('Restarting worker:', worker_name)
        self._workers[worker_name]['pid'] = self._start_worker(worker_name, self._workers[worker_name]['gpu_index'])

    def _get_available_gpus(self):
        return {str(gpu.id): {'gpu_index': gpu.id,
                              'free_gpu_memory': gpu.memoryFree} for gpu in GPUtil.getGPUs()}

    def _send_heartbeat_to_server(self, first=False):
        self._management_q.put(('CLIENT_HEARTBEAT', {'name': self._name, 'first': first,
                                                     'worker_names': list(self._workers)}))

    def _kill_all_workers(self):
        for worker in self._workers:
            self._kill_worker(worker)

        self._workers = {}

    def _kill_worker(self, worker_name):
        if worker_name in self._workers:
            psutil.Process(self._workers[worker_name]['pid']).send_signal(signal.SIGKILL)

    def _start_worker(self, worker_name, gpu_index):

        worker_q_name = f'{self._name}@{worker_name}'

        worker_process = ExperimentWorkerExecutor(worker_name, self._name, gpu_index, self._jobs_q_name,
                                                  self._results_q_name, MANAGEMENT_Q_NAME, worker_q_name,
                                                  self._worker_class, self._init_data, self._host, self._port,
                                                  self._password)
        worker_process.start()
        return worker_process.pid

    def _handle_task_assignment(self, data):

        self._kill_all_workers()

        self._jobs_q_name = data['jobs_q_name']
        self._results_q_name = data['results_q_name']
        self._worker_class = data['worker_class']
        self._init_data = data['init_data']
        self._gpu_memory_required = data['gpu_memory_required']
        self._task_assigned = True

        self._start_available_workers()

    def stop(self):
        self._kill_all_workers()




import pickle
import socket
import time

import redis

# https://github.com/andymccurdy/redis-py/issues/722
from distributed_computing.globals import DEFAULT_PORT


def safe_block(foo):
    def safe_foo(self, *args, **kwargs):

        while True:
            try:
                return foo(self, *args, **kwargs)
            except (socket.timeout, redis.exceptions.TimeoutError, redis.exceptions.ConnectionError) as e:
                time.sleep(3)
                self._connect()

    return safe_foo


class RedisQueue(object):
    """Simple Queue with Redis Backend"""
    def __init__(self, name, password, host='localhost', port=DEFAULT_PORT, namespace='queue'):

        self._host = host
        self._port = port
        self._password = password

        self._connect()
        self.key = '%s:%s' %(namespace, name)

    def _connect(self):
        self.__db = redis.Redis(host=self._host, port=self._port, password=self._password, socket_keepalive=True,
                                socket_timeout=60)

    def reconnect(self):
        self._connect()

    @safe_block
    def clear(self):
        self.__db.delete(self.key)
        return self

    @safe_block
    def qsize(self):
        """Return the approximate size of the queue."""
        return self.__db.llen(self.key)

    @safe_block
    def empty(self):
        """Return True if the queue is empty, False otherwise."""
        return self.qsize() == 0

    @safe_block
    def put(self, item):
        """Put item into the queue."""
        self.__db.rpush(self.key, pickle.dumps(item))

    @safe_block
    def get(self, block=True, timeout=None):
        """Remove and return an item from the queue.

        If optional args block is true and timeout is None (the default), block
        if necessary until an item is available."""
        if block:
            item = self.__db.blpop(self.key, timeout=timeout)

            if item:
                item = item[1]
        else:
            item = self.__db.lpop(self.key)

        if item is not None:
            return pickle.loads(item)

    def get_nowait(self):
        """Equivalent to get(False)."""
        return self.get(False)

    def get_all_nowait(self):

        messages = []
        message = self.get(False)

        while message is not None:
            messages.append(message)
            message = self.get(False)

        return messages



import multiprocessing
import time

from distributed_computing.api import start_head_node, WorkerInterface, Pool

PASSWORD = '9976145'


class TestWorker(WorkerInterface):

    def __init__(self, arg, kwarg=None):
        self._arg = arg
        self._kwarg = kwarg
        print('Initialization done')

    def run(self, arg, kwarg=None):
        print('Running for', arg, 'seconds')
        time.sleep(arg)
        print('Done running')
        return kwarg

    def handle_update(self, arg, kwarg=None):
        print('Update received')


if __name__ == '__main__':

    multiprocessing.set_start_method('spawn')

    print('######## Starting head node')
    terminate, wait = start_head_node(PASSWORD)

    print('######## Done starting head node')
    init_data = {'args': [1], 'kwargs': {'kwarg': 2}}

    print('######## Creating pool')
    pool = Pool(TestWorker, init_data, PASSWORD)

    data = [{'args': [i], 'kwargs': {'kwarg': i}} for i in range(30)]

    print('######## Processing data')
    for result in pool.imap_unordered(data):
        print('Result:', result)
        pool.update_workers({'args': [result], 'kwargs': {'kwarg': result}})

    print('######## Terminating')
    terminate()




from setuptools import setup

setup(name='distributed_computing',
      version='0.2.0',
      description='Distributed GPU computing with python',
      packages=['distributed_computing'],
      author_email='boaz85@gmail.com',
      install_requires=['redis-server==5.0.7', 'GPUtil==1.4.0', 'psutil==5.7.3', 'setproctitle==1.1.10',
                        'redis==3.5.3', 'requests==2.23.0', 'service_identity==18.1.0'],
      scripts=['distributed_computing/start_node'],
      zip_safe=False)
